#define VERSION		"4.01 Beta"

/* Bayesian Analysis in Molecular Biology and Evolution (BAMBE) version 4.01
 * Beta
 * (c) Copyright 1999, 2001, 2002, 2003, Donald Simon & Bret Larget
 * Department of Mathematics/Computer Science, Duquesne University
 */

#include "util.h"
#include "parameters.h"
#include "runsettings.h"
#include "site.h"
#include "basedata.h"
#include "sitedata.h"
#include "tree.h"
#include "files.h"

#include <iostream>
#include <iomanip>

#define REROOT_PROB	.01

void updateIntervals(RunSettings& rs, const Tree& t, int cycle, 
		     const Parameters& params, AcceptCounts& ac, 
		     Files& files, int tuning);
		     
void global(RunSettings& rs, const SiteData& sd, Parameters& params,
	    Tree& t, Files& files, AcceptCounts& ac, int burn);
	    
void local(RunSettings& rs, const SiteData& sd, Parameters& params,
	   Tree& t, Files& files, AcceptCounts& ac, int burn);

void printInitialInformation(Files& f, const RunSettings& rs,
			     const SiteData& sd, const Parameters& params);
			     	   
void updateIntervals(RunSettings& rs, const Tree& t, int cycle, 
		     const Parameters& params, AcceptCounts& ac, 
		     Files& files, int tuning)
{
  if(cycle % rs.getSampleInterval() == 0) {
    t.printTopology(files.ftop);
    files.ftop << endl;
    if(rs.getPrintAllTrees())
      t.printNumberTree(files.ftree);
    params.print(files.fkappa,rs,cycle,ac.accepts,ac.kaccepts,ac.paccepts,
		 t.getLoglikelihood(),t.getRadius(),
		 rs.getSampleInterval());
    ac.accepts = ac.kaccepts = ac.paccepts = 0;
    files.flogp << setiosflags(ios::showpoint | ios::fixed) << setw(10) 
		<< setprecision(6) << t.getLoglikelihood() << endl;
  }
  if(tuning && (cycle % rs.getTuneInterval() == 0)
     && ((double)(ac.accept)/(double)(rs.getTuneInterval()) < .15))
    rs.halveValleyWin();
  if(cycle % rs.getWindowInterval() == 0) {
    t.printTopology(cout);
    cout << endl;
    params.printToScreen(cout,rs,cycle,ac.accept,ac.kaccept,ac.paccept,
		 t.getLoglikelihood(),t.getRadius(),
		 rs.getWindowInterval());
    ac.accept = ac.kaccept = ac.paccept = 0;
  }
}

void global(RunSettings& rs, const SiteData& sd, Parameters& params,
	    Tree& t, Files& files, AcceptCounts& ac, int burn)
{
  int numCycles = (burn ? rs.getBurn() : rs.getNumCycles());
  for(int cycle=1; cycle <= numCycles; cycle++) {
    if(!burn)
      t.checkParamUpdate(rs,sd,params,cycle,ac);
    if(!rs.getMclock() && Rand::runif() < REROOT_PROB) {
      t.reroot(sd,params);
      ac.accept++;
      ac.accepts++;
    }
    else {
      t.saveGlobal();
      t.globalUpdate(rs,sd,params);
      if(t.getRadius() < 100 &&
	 acceptable(t.getLoglikelihood(),t.getSavedLoglikelihood(),1.0)) {
	ac.accept++;
	ac.accepts++;
      } 
      else
	t.restoreGlobal();
    }
    updateIntervals(rs,t,cycle,params,ac,files,burn);
  }
}

void local(RunSettings& rs, const SiteData& sd, Parameters& params,
	   Tree& t, Files& files, AcceptCounts& ac, int burn)
{
  int numCycles = (burn ? rs.getBurn() : rs.getNumCycles());
  for(int cycle=1; cycle <= numCycles; cycle++) {
    if(!burn)
      t.checkParamUpdate(rs,sd,params,cycle,ac);
    t.localUpdate(rs,sd,params);
    if(t.getRadius() < 100 &&
       acceptable(t.getLoglikelihood(),t.getSavedLoglikelihood(),
		  t.getHastingsRatio())) {
      ac.accept++;
      ac.accepts++;
    }
    else
      t.restoreLocal();
    updateIntervals(rs,t,cycle,params,ac,files,burn);
  }
}

void printInitialInformation(Files& f, const RunSettings& rs,
			     const SiteData& sd, const Parameters& params)
{
  /* Prints out summary information. */

  f.fout << "Output generated by BAMBE version " << VERSION << endl;
  f.fout << "Data Summary" << endl << endl;
  sd.printInfo(f.fout,rs);
  params.printInfo(f.fout,rs,sd);
  sd.printMoreInfo(f.fout);
  rs.print(f.fout);
}

int main(void)
{
  cout << "Reading Runtime Parameters...." << endl;
  RunSettings rs(cin);			// Read in the run settings.
  cout << "Opening Files...." << endl;
  Files files(rs);
  Rand(rs.getSeed());			// Initialize the random number gen.
  cout << "Reading Data file...." << endl;
  BaseData bd(rs);	 		// Read in the base data.
  Site(bd.getNumTaxa());	        // Set the length of all sites.
  cout << "Creating initial parameters...." << endl;
  SiteData sd(rs,bd);			// Generate the site data.
  Parameters params(rs,sd);		// Set up the parameters.
  cout << "Writing initial summary information to file...." << endl;
  printInitialInformation(files,rs,sd,params);
  cout << "Creating initial tree...." << endl;
  Tree t(rs,sd,params);     		// Create or read in initial tree.
  AcceptCounts ac;

  t.print(cout);
  if(rs.getNewickFormat())
    cout << "log loglikelihood = " << t.getLoglikelihood() << endl << endl;
  t.printTopology(cout);
  cout << endl;
  cout << "Starting burn in...." << endl;
  switch (rs.getBurnAlgorithm()) {
  case RSAlgorithm::GLOBAL : global(rs,sd,params,t,files,ac,1); break;
  case RSAlgorithm::LOCAL  : local(rs,sd,params,t,files,ac,1); break;
  }

  cout << "Entering main loop...." << endl;
  switch (rs.getMainAlgorithm()) {
  case RSAlgorithm::GLOBAL: global(rs,sd,params,t,files,ac,0); break;
  case RSAlgorithm::LOCAL:  local(rs,sd,params,t,files,ac,0); break;
  }

  t.printTopology(cout);
  cout << endl;
  t.print(files.flast);
  t.print(cout);
  if(rs.getNewickFormat())
    cout << "log loglikelihood = " << t.getLoglikelihood() << endl << endl;
  return 0;
}


